/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_scrn.c -- master for refresh, status bar, console, chat, notify, etc

/*

  full screen console
  put up loading plaque
  blanked background with loading plaque
  blanked background with menu
  cinematics
  full screen image for quit and victory

  end of unit intermissions

  */

#include "client.h"

float		scr_con_current;	// aproaches scr_conlines at scr_conspeed
float		scr_conlines;		// 0.0 to 1.0 lines of console to display

qboolean	scr_initialized;		// ready to draw

int			scr_draw_loading;

vrect_t		scr_vrect;		// position of render window on screen

cvar_t		*scr_conspeed;
cvar_t		*scr_centertime;
cvar_t		*scr_showturtle;
cvar_t		*scr_showpause;
cvar_t		*scr_printspeed;

cvar_t		*scr_netgraph;
cvar_t		*scr_timegraph;
cvar_t		*scr_debuggraph;
cvar_t		*scr_graphheight;
cvar_t		*scr_graphscale;
cvar_t		*scr_graphshift;
cvar_t		*scr_drawall;
cvar_t		*scr_showface;

typedef struct
{
	int		x1, y1, x2, y2;
} dirty_t;

dirty_t		scr_dirty, scr_old_dirty[2];

char		crosshair_pic[MAX_QPATH];
int			crosshair_width, crosshair_height;

void SCR_TimeRefresh_f (void);
void SCR_Loading_f (void);

// standard QII architecture makes testing for a lost device problematic, so we need a separate action for it
qboolean GLimp_TestDevice (void);


/*
===============================================================================

BAR GRAPHS

===============================================================================
*/

/*
==============
CL_AddNetgraph

A new packet was just parsed
==============
*/
void CL_AddNetgraph (void)
{
	int		i;
	int		in;
	int		ping;

	// if using the debuggraph for something else, don't
	// add the net lines
	if (scr_debuggraph->value || scr_timegraph->value)
		return;

	for (i = 0; i < cls.netchan.dropped; i++)
		SCR_DebugGraph (30, 0x40);

	for (i = 0; i < cl.surpressCount; i++)
		SCR_DebugGraph (30, 0xdf);

	// see what the latency was on this packet
	in = cls.netchan.incoming_acknowledged & (CMD_BACKUP - 1);
	ping = cls.realtime - cl.cmd_time[in];
	ping /= 30;

	if (ping > 30)
		ping = 30;

	SCR_DebugGraph (ping, 0xd0);
}


typedef struct
{
	float	value;
	int		color;
} graphsamp_t;

static	int			current;
static	graphsamp_t	values[1024];

/*
==============
SCR_DebugGraph
==============
*/
void SCR_DebugGraph (float value, int color)
{
	values[current&1023].value = value;
	values[current&1023].color = color;
	current++;
}

/*
==============
SCR_DrawDebugGraph
==============
*/
void SCR_DrawDebugGraph (void)
{
	int		a, x, y, w, i, h;
	float	v;
	int		color;

	// draw the graph
	w = scr_vrect.width;

	x = scr_vrect.x;
	y = scr_vrect.y + scr_vrect.height;
	Draw_Fill (x, y - scr_graphheight->value,
			   w, scr_graphheight->value, 8);

	for (a = 0; a < w; a++)
	{
		i = (current - 1 - a + 1024) & 1023;
		v = values[i].value;
		color = values[i].color;
		v = v * scr_graphscale->value + scr_graphshift->value;

		if (v < 0)
			v += scr_graphheight->value * (1 + (int) (-v / scr_graphheight->value));

		h = (int) v % (int) scr_graphheight->value;
		Draw_Fill (x + w - 1 - a, y - h, 1,	h, color);
	}
}

/*
===============================================================================

CENTER PRINTING

===============================================================================
*/

char		scr_centerstring[1024];
float		scr_centertime_start;	// for slow victory printing
float		scr_centertime_off;
int			scr_center_lines;
int			scr_erase_center;

/*
==============
SCR_CenterPrint

Called for important messages that should stay in the center of the screen
for a few moments
==============
*/
void SCR_CenterPrint (char *str)
{
	char	*s;
	char	line[64];
	int		i, j, l;

	strncpy (scr_centerstring, str, sizeof (scr_centerstring) - 1);
	scr_centertime_off = scr_centertime->value;
	scr_centertime_start = cl.time;

	// count the number of lines for centering
	scr_center_lines = 1;
	s = str;

	while (*s)
	{
		if (*s == '\n')
			scr_center_lines++;

		s++;
	}

	// echo it to the console
	Com_Printf ("\n\n\35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n\n");

	s = str;

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
			if (s[l] == '\n' || !s[l])
				break;

		for (i = 0; i < (40 - l) / 2; i++)
			line[i] = ' ';

		for (j = 0; j < l; j++)
		{
			line[i++] = s[j];
		}

		line[i] = '\n';
		line[i+1] = 0;

		Com_Printf ("%s", line);

		while (*s && *s != '\n')
			s++;

		if (!*s)
			break;

		s++;		// skip the \n
#pragma warning (disable: 4127)
		// suppress conditional expression is constant so that we can use this construct
	} while (1);
#pragma warning (1: 4127)

	Com_Printf ("\n\n\35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n\n");
	Con_ClearNotify ();
}


void SCR_DrawCenterString (void)
{
	char	*start;
	int		l;
	int		j;
	int		x, y;
	int		remaining;

	// the finale prints the characters one at a time
	remaining = 9999;

	scr_erase_center = 0;
	start = scr_centerstring;

	// always center properly
	y = (d3d_2DMode.Height * 0.35) - (scr_center_lines * 4);

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
			if (start[l] == '\n' || !start[l])
				break;

		x = (d3d_2DMode.Width - l * 8) / 2;

		for (j = 0; j < l; j++, x += 8)
		{
			Draw_Char (x, y, start[j]);

			if (!remaining--)
				return;
		}

		y += 8;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;

		start++;		// skip the \n
#pragma warning (disable: 4127)
		// suppress conditional expression is constant so that we can use this construct
	} while (1);
#pragma warning (1: 4127)
}

void SCR_CheckDrawCenterString (void)
{
	scr_centertime_off -= cls.frametime;

	if (scr_centertime_off <= 0)
		return;

	SCR_DrawCenterString ();
}

//=============================================================================

/*
=================
SCR_CalcVrect

Sets scr_vrect, the coordinates of the rendered window
=================
*/
static void SCR_CalcVrect (void)
{
	scr_vrect.x = 0;
	scr_vrect.y = 0;
	scr_vrect.width = viddef.Width;
	scr_vrect.height = viddef.Height;
}


/*
=================
SCR_Sky_f

Set a specific sky and rotation speed
=================
*/
void SCR_Sky_f (void)
{
	float	rotate;
	vec3_t	axis;

	if (Cmd_Argc () < 2)
	{
		Com_Printf ("Usage: sky <basename> <rotate> <axis x y z>\n");
		return;
	}

	if (Cmd_Argc () > 2)
		rotate = atof (Cmd_Argv (2));
	else
		rotate = 0;

	if (Cmd_Argc () == 6)
	{
		axis[0] = atof (Cmd_Argv (3));
		axis[1] = atof (Cmd_Argv (4));
		axis[2] = atof (Cmd_Argv (5));
	}
	else
	{
		axis[0] = 0;
		axis[1] = 0;
		axis[2] = 1;
	}

	R_SetSky (Cmd_Argv (1), rotate, axis);
}

//============================================================================

/*
==================
SCR_Init
==================
*/
void SCR_Init (void)
{
	scr_conspeed = Cvar_Get ("scr_conspeed", "3", 0);
	scr_showturtle = Cvar_Get ("scr_showturtle", "0", 0);
	scr_showpause = Cvar_Get ("scr_showpause", "1", 0);
	scr_centertime = Cvar_Get ("scr_centertime", "2.5", 0);
	scr_printspeed = Cvar_Get ("scr_printspeed", "8", 0);
	scr_netgraph = Cvar_Get ("netgraph", "0", 0);
	scr_timegraph = Cvar_Get ("timegraph", "0", 0);
	scr_debuggraph = Cvar_Get ("debuggraph", "0", 0);
	scr_graphheight = Cvar_Get ("graphheight", "32", 0);
	scr_graphscale = Cvar_Get ("graphscale", "1", 0);
	scr_graphshift = Cvar_Get ("graphshift", "0", 0);
	scr_drawall = Cvar_Get ("scr_drawall", "0", 0);
	scr_showface = Cvar_Get ("scr_showface", "0", CVAR_ARCHIVE);

	// register our commands
	Cmd_AddCommand ("timerefresh", SCR_TimeRefresh_f);
	Cmd_AddCommand ("loading", SCR_Loading_f);
	Cmd_AddCommand ("sky", SCR_Sky_f);

	scr_initialized = true;
}


/*
==============
SCR_DrawNet
==============
*/
void SCR_DrawNet (void)
{
	if (cls.netchan.outgoing_sequence - cls.netchan.incoming_acknowledged
			< CMD_BACKUP - 1)
		return;

	Draw_Pic (scr_vrect.x + 64, scr_vrect.y, "net");
}

/*
==============
SCR_DrawPause
==============
*/
void SCR_DrawPause (void)
{
	int		w, h;

	if (!scr_showpause->value)		// turn off for screenshots
		return;

	if (!cl_paused->value)
		return;

	Draw_GetPicSize (&w, &h, "pause");
	Draw_Pic ((d3d_2DMode.Width - w) / 2, d3d_2DMode.Height / 2 + 8, "pause");
}

/*
==============
SCR_DrawLoading
==============
*/
void SCR_DrawLoading (void)
{
	int		w, h;

	if (!scr_draw_loading)
		return;

	if (scr_draw_loading == 2)
		Draw_Fill (0, 0, d3d_2DMode.Width, d3d_2DMode.Height, 0);

	scr_draw_loading = 0;

	Draw_GetPicSize (&w, &h, "loading");
	Draw_Pic ((d3d_2DMode.Width - w) / 2, (d3d_2DMode.Height - h) / 2, "loading");
}

//=============================================================================

/*
==================
SCR_RunConsole

Scroll it up or down
==================
*/
void SCR_RunConsole (void)
{
	static int oldrealtime = cls.realtime;

	float scr_contime = (cls.realtime - oldrealtime) * 0.001f;
	oldrealtime = cls.realtime;

	// decide on the height of the console
	if (cls.key_dest == key_console)
		scr_conlines = 0.5;		// half screen
	else scr_conlines = 0;		// none visible

	if (scr_conlines < scr_con_current)
	{
		scr_con_current -= scr_conspeed->value * scr_contime;

		if (scr_conlines > scr_con_current)
			scr_con_current = scr_conlines;

	}
	else if (scr_conlines > scr_con_current)
	{
		scr_con_current += scr_conspeed->value * scr_contime;

		if (scr_conlines < scr_con_current)
			scr_con_current = scr_conlines;
	}

}

/*
==================
SCR_DrawConsole
==================
*/
void SCR_DrawConsole (void)
{
	Con_CheckResize ();

	if (cls.state == ca_disconnected || cls.state == ca_connecting)
	{
		// forced full screen console
		Con_DrawConsole (1.0f, 1.0f);
		return;
	}

	if (cls.state != ca_active || !cl.refresh_prepped)
	{
		// connected, but can't render
		Con_DrawConsole (0.5f, 1.0f);
		Draw_Fill (0, d3d_2DMode.Height / 2, d3d_2DMode.Width, d3d_2DMode.Height / 2, 0);
		return;
	}

	if (scr_con_current)
	{
		Con_DrawConsole (scr_con_current, scr_con_current);
	}
	else
	{
		if (cls.key_dest == key_game || cls.key_dest == key_message)
			Con_DrawNotify ();	// only draw notify in game
	}
}

//=============================================================================

/*
================
SCR_BeginLoadingPlaque
================
*/
void SCR_BeginLoadingPlaque (void)
{
	S_StopAllSounds ();
	cl.sound_prepped = false;		// don't play ambients
	CDAudio_Stop ();

	if (cls.disable_screen)
		return;

	if (developer->value)
		return;

	if (cls.state == ca_disconnected)
		return;	// if at console, don't bring up the plaque

	if (cls.key_dest == key_console)
		return;

	if (cl.cinematictime > 0)
		scr_draw_loading = 2;	// clear to black first
	else scr_draw_loading = 1;

	SCR_UpdateScreen ();
	cls.disable_screen = Sys_Milliseconds ();
	cls.disable_servercount = cl.servercount;
}

/*
================
SCR_EndLoadingPlaque
================
*/
void SCR_EndLoadingPlaque (void)
{
	cls.disable_screen = 0;
	Con_ClearNotify ();
}

/*
================
SCR_Loading_f
================
*/
void SCR_Loading_f (void)
{
	SCR_BeginLoadingPlaque ();
}

/*
================
SCR_TimeRefresh_f
================
*/
int entitycmpfnc (const entity_t *a, const entity_t *b)
{
	/*
	** all other models are sorted by model then skin
	*/
	if (a->model == b->model)
	{
		return ((int) a->skin - (int) b->skin);
	}
	else
	{
		return ((int) a->model - (int) b->model);
	}
}

void SCR_TimeRefresh_f (void)
{
	int		i;
	int		start, stop;
	float	time;

	if (cls.state != ca_active)
		return;

	start = Sys_Milliseconds ();

	if (Cmd_Argc () == 2)
	{
		// run without page flipping
		R_BeginFrame (0);

		for (i = 0; i < 128; i++)
		{
			cl.refdef.viewangles[1] = i / 128.0 * 360.0;
			R_RenderFrame (&cl.refdef);
		}

		GLimp_EndFrame ();
	}
	else
	{
		for (i = 0; i < 128; i++)
		{
			cl.refdef.viewangles[1] = i / 128.0 * 360.0;

			R_BeginFrame (0);
			R_RenderFrame (&cl.refdef);
			GLimp_EndFrame ();
		}
	}

	stop = Sys_Milliseconds ();
	time = (stop - start) / 1000.0;
	Com_Printf ("%f seconds (%f fps)\n", time, 128 / time);
}


/*
==============
SCR_TileClear

Clear any parts of the tiled background that were drawn on last frame
==============
*/
void SCR_TileClear (void)
{
}


//===============================================================


#define STAT_MINUS		10	// num frame for '-' stats digit
char *sb_nums[2][11] =
{
	{
		"num_0", "num_1", "num_2", "num_3", "num_4", "num_5",
		"num_6", "num_7", "num_8", "num_9", "num_minus"
	},
	{
		"anum_0", "anum_1", "anum_2", "anum_3", "anum_4", "anum_5",
		"anum_6", "anum_7", "anum_8", "anum_9", "anum_minus"
	}
};

#define	ICON_WIDTH	24
#define	ICON_HEIGHT	24
#define	CHAR_WIDTH	16
#define	ICON_SPACE	8



/*
================
SizeHUDString

Allow embedded \n in the string
================
*/
void SizeHUDString (char *string, int *w, int *h)
{
	int		lines, width, s_current;

	lines = 1;
	width = 0;

	s_current = 0;

	while (*string)
	{
		if (*string == '\n')
		{
			lines++;
			s_current = 0;
		}
		else
		{
			s_current++;

			if (s_current > width)
				width = s_current;
		}

		string++;
	}

	*w = width * 8;
	*h = lines * 8;
}

void DrawHUDString (char *string, int x, int y, int centerwidth, int xor)
{
	int		margin;
	char	line[1024];
	int		width;
	int		i;

	margin = x;

	while (*string)
	{
		// scan out one line of text from the string
		width = 0;

		while (*string && *string != '\n')
			line[width++] = *string++;

		line[width] = 0;

		if (centerwidth)
			x = margin + (centerwidth - width * 8) / 2;
		else
			x = margin;

		for (i = 0; i < width; i++)
		{
			Draw_Char (x, y, line[i] ^ xor);
			x += 8;
		}

		if (*string)
		{
			string++;	// skip the \n
			x = margin;
			y += 8;
		}
	}
}


/*
==============
SCR_DrawField
==============
*/
void SCR_DrawField (int x, int y, int color, int width, int value)
{
	char	num[16], *ptr;
	int		l;
	int		frame;

	if (width < 1)
		return;

	// draw number string
	if (width > 5)
		width = 5;

	Com_sprintf (num, sizeof (num), "%i", value);
	l = strlen (num);

	if (l > width)
		l = width;

	x += 2 + CHAR_WIDTH * (width - l);

	ptr = num;

	while (*ptr && l)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr - '0';

		Draw_Pic (x, y, sb_nums[color][frame]);
		x += CHAR_WIDTH;
		ptr++;
		l--;
	}
}


/*
===============
SCR_TouchPics

Allows rendering code to cache all needed sbar graphics
===============
*/
void SCR_TouchPics (void)
{
	int		i, j;

	for (i = 0; i < 2; i++)
		for (j = 0; j < 11; j++)
			Draw_FindPic (sb_nums[i][j]);

	if (crosshair->value)
	{
		// fixme - should this be a Cvar_SetValue?
		if (crosshair->value < 0) crosshair->value = 0;
		if (crosshair->value > 3) crosshair->value = 3;

		Com_sprintf (crosshair_pic, sizeof (crosshair_pic), "ch%i", (int) (crosshair->value));
		Draw_GetPicSize (&crosshair_width, &crosshair_height, crosshair_pic);

		if (!crosshair_width) crosshair_pic[0] = 0;
	}
}


/*
================
SCR_ExecuteLayoutString

================
*/
void SCR_ExecuteLayoutString (char *s)
{
	int		x, y;
	int		value;
	char	*token;
	int		width;
	int		index;
	clientinfo_t	*ci;

	static int lastitem = -1;
	static int itemtime = 0;

	if (cls.state != ca_active || !cl.refresh_prepped)
		return;

	if (!s[0])
		return;

	x = 0;
	y = 0;
	width = 3;

	while (s)
	{
		token = COM_Parse (&s);

		if (!strcmp (token, "xl"))
		{
			token = COM_Parse (&s);
			x = atoi (token);
			continue;
		}

		if (!strcmp (token, "xr"))
		{
			token = COM_Parse (&s);
			x = d3d_2DMode.Width + atoi (token);
			continue;
		}

		if (!strcmp (token, "xv"))
		{
			token = COM_Parse (&s);
			x = d3d_2DMode.Width / 2 - 160 + atoi (token);
			continue;
		}

		if (!strcmp (token, "yt"))
		{
			token = COM_Parse (&s);
			y = atoi (token);
			continue;
		}

		if (!strcmp (token, "yb"))
		{
			token = COM_Parse (&s);
			// hacky hacky hacky pooh - offset the y by 4 from the bottom so it ain't flush
			y = d3d_2DMode.Height + atoi (token) - 4;
			continue;
		}

		if (!strcmp (token, "yv"))
		{
			token = COM_Parse (&s);
			y = d3d_2DMode.Height / 2 - 120 + atoi (token);
			continue;
		}

		if (!strcmp (token, "pic"))
		{
			// draw a pic from a stat number
			char *statpic = NULL;
			static char facepic[256];
			int statnum;

			token = COM_Parse (&s);
			statnum = atoi (token);

			value = cl.frame.playerstate.stats[statnum];

			if (value >= MAX_IMAGES) Com_Error (ERR_DROP, "Pic >= MAX_IMAGES");

			// this is to show the face icon for the selected skin in the sbar instead of the regular health icon
			if (statnum == STAT_HEALTH_ICON && scr_showface->value)
			{
				extern cvar_t *skin;
				int w = 24;
				int h = 24;

				// check first to see if the pic exists
				sprintf (facepic, "/players/%s_i.pcx", skin->string);
				Draw_GetPicSize (&w, &h, facepic);

				if (w > 0 && h > 0)
				{
					// draw it at the correct scaled size for the HUD health pic
					// (which may not be 24 x 24)
					statpic = cl.configstrings[CS_IMAGES + value];

					if (statpic)
					{
						Draw_GetPicSize (&w, &h, statpic);

						if (w < 0) w = 24;

						if (h < 0) h = 24;
					}

					Draw_Pic (x, y, w, h, facepic);
					continue;
				}
			}

			// fallback to default
			statpic = cl.configstrings[CS_IMAGES + value];

			if (statpic) Draw_Pic (x, y, statpic);

			if (statnum == STAT_SELECTED_ICON)
			{
				if (cl.frame.playerstate.stats[STAT_SELECTED_ITEM] != lastitem)
				{
					itemtime = cl.time + 1500;
					lastitem = cl.frame.playerstate.stats[STAT_SELECTED_ITEM];
				}

				if (cl.time < itemtime)
				{
					int i;
					int xx;

					for (i = 0, xx = x + 32; ; i++, xx += 8)
					{
						if (!cl.configstrings[CS_ITEMS + cl.frame.playerstate.stats[STAT_SELECTED_ITEM]][i]) break;

						Draw_Char (xx, y + 8, cl.configstrings[CS_ITEMS + cl.frame.playerstate.stats[STAT_SELECTED_ITEM]][i]);
					}
				}
			}

			continue;
		}

		if (!strcmp (token, "client"))
		{
			// draw a deathmatch client block
			int		score, ping, time;

			token = COM_Parse (&s);
			x = d3d_2DMode.Width / 2 - 160 + atoi (token);
			token = COM_Parse (&s);
			y = d3d_2DMode.Height / 2 - 120 + atoi (token);

			token = COM_Parse (&s);
			value = atoi (token);

			if (value >= MAX_CLIENTS || value < 0)
				Com_Error (ERR_DROP, "client >= MAX_CLIENTS");

			ci = &cl.clientinfo[value];

			token = COM_Parse (&s);
			score = atoi (token);

			token = COM_Parse (&s);
			ping = atoi (token);

			token = COM_Parse (&s);
			time = atoi (token);

			DrawAltString (x + 32, y, ci->name);
			DrawString (x + 32, y + 8,  "Score: ");
			DrawAltString (x + 32 + 7 * 8, y + 8,  va ("%i", score));
			DrawString (x + 32, y + 16, va ("Ping:  %i", ping));
			DrawString (x + 32, y + 24, va ("Time:  %i", time));

			if (!ci->icon)
				ci = &cl.baseclientinfo;

			Draw_Pic (x, y, ci->iconname);
			continue;
		}

		if (!strcmp (token, "ctf"))
		{
			// draw a ctf client block
			int		score, ping;
			char	block[80];

			token = COM_Parse (&s);
			x = d3d_2DMode.Width / 2 - 160 + atoi (token);
			token = COM_Parse (&s);
			y = d3d_2DMode.Height / 2 - 120 + atoi (token);

			token = COM_Parse (&s);
			value = atoi (token);

			if (value >= MAX_CLIENTS || value < 0)
				Com_Error (ERR_DROP, "client >= MAX_CLIENTS");

			ci = &cl.clientinfo[value];

			token = COM_Parse (&s);
			score = atoi (token);

			token = COM_Parse (&s);
			ping = atoi (token);

			if (ping > 999)
				ping = 999;

			sprintf (block, "%3d %3d %-12.12s", score, ping, ci->name);

			if (value == cl.playernum)
				DrawAltString (x, y, block);
			else
				DrawString (x, y, block);

			continue;
		}

		if (!strcmp (token, "picn"))
		{
			// draw a pic from a name
			token = COM_Parse (&s);
			Draw_Pic (x, y, token);
			continue;
		}

		if (!strcmp (token, "num"))
		{
			// draw a number
			token = COM_Parse (&s);
			width = atoi (token);
			token = COM_Parse (&s);
			value = cl.frame.playerstate.stats[atoi (token)];

			// another hack - offset the x by 4
			SCR_DrawField (x - 4, y, 0, width, value);
			continue;
		}

		if (!strcmp (token, "hnum"))
		{
			// health number
			int		color;

			width = 3;
			value = cl.frame.playerstate.stats[STAT_HEALTH];

			if (value > 25)
				color = 0;	// green
			else if (value > 0)
				color = (cl.frame.serverframe >> 2) & 1;		// flash
			else
				color = 1;

			if (cl.frame.playerstate.stats[STAT_FLASHES] & 1)
				Draw_Pic (x - 4, y, "field_3");

			SCR_DrawField (x - 4, y, color, width, value);
			continue;
		}

		if (!strcmp (token, "anum"))
		{
			// ammo number
			int		color;

			width = 3;
			value = cl.frame.playerstate.stats[STAT_AMMO];

			if (value > 5)
				color = 0;	// green
			else if (value >= 0)
				color = (cl.frame.serverframe >> 2) & 1;		// flash
			else
				continue;	// negative number = don't show

			if (cl.frame.playerstate.stats[STAT_FLASHES] & 4)
				Draw_Pic (x - 4, y, "field_3");

			SCR_DrawField (x - 4, y, color, width, value);
			continue;
		}

		if (!strcmp (token, "rnum"))
		{
			// armor number
			int		color;

			width = 3;
			value = cl.frame.playerstate.stats[STAT_ARMOR];

			if (value < 1)
				continue;

			color = 0;	// green

			if (cl.frame.playerstate.stats[STAT_FLASHES] & 2)
				Draw_Pic (x - 4, y, "field_3");

			SCR_DrawField (x - 4, y, color, width, value);
			continue;
		}


		if (!strcmp (token, "stat_string"))
		{
			token = COM_Parse (&s);
			index = atoi (token);

			if (index < 0 || index >= MAX_CONFIGSTRINGS)
				Com_Error (ERR_DROP, "Bad stat_string index");

			index = cl.frame.playerstate.stats[index];

			if (index < 0 || index >= MAX_CONFIGSTRINGS)
				Com_Error (ERR_DROP, "Bad stat_string index");

			DrawString (x, y, cl.configstrings[index]);
			continue;
		}

		if (!strcmp (token, "cstring"))
		{
			token = COM_Parse (&s);
			DrawHUDString (token, x, y, 320, 0);
			continue;
		}

		if (!strcmp (token, "string"))
		{
			token = COM_Parse (&s);
			DrawString (x, y, token);
			continue;
		}

		if (!strcmp (token, "cstring2"))
		{
			token = COM_Parse (&s);
			DrawHUDString (token, x, y, 320, 0x80);
			continue;
		}

		if (!strcmp (token, "string2"))
		{
			token = COM_Parse (&s);
			DrawAltString (x, y, token);
			continue;
		}

		if (!strcmp (token, "if"))
		{
			// draw a number
			token = COM_Parse (&s);
			value = cl.frame.playerstate.stats[atoi (token)];

			if (!value)
			{
				// skip to endif
				while (s && strcmp (token, "endif"))
				{
					token = COM_Parse (&s);
				}
			}

			continue;
		}
	}
}


/*
================
SCR_DrawStats

The status bar is a small layout program that
is based on the stats array
================
*/
void SCR_DrawStats (void)
{
	SCR_ExecuteLayoutString (cl.configstrings[CS_STATUSBAR]);
}


/*
================
SCR_DrawLayout

================
*/
#define	STAT_LAYOUTS		13

void SCR_DrawLayout (void)
{
	if (!cl.frame.playerstate.stats[STAT_LAYOUTS])
		return;

	SCR_ExecuteLayoutString (cl.layout);
}

//=======================================================

/*
==================
SCR_UpdateScreen

This is called every frame, and can also be called explicitly to flush
text to the screen.
==================
*/
void SCR_DisableCinematicPalette (void)
{
	if (cl.cinematicpalette_active)
	{
		R_SetPalette (NULL);
		cl.cinematicpalette_active = false;
	}
}


void SCR_UpdateScreen (void)
{
	// if the screen is disabled (loading plaque is up, or vid mode changing)
	// do nothing at all
	if (cls.disable_screen)
	{
		if (Sys_Milliseconds() - cls.disable_screen > 120000)
		{
			cls.disable_screen = 0;
			Com_Printf ("Loading plaque timed out.\n");
		}

		return;
	}

	if (!scr_initialized || !con.initialized)
		return;				// not initialized yet

	// if we lost the device we draw nothing
	if (!GLimp_TestDevice ()) return;

	// figure the 2D ui scale early
	extern cvar_t *r_uiscale;

	if (r_uiscale->value > 1)
	{
		d3d_2DMode.Width = (float) viddef.Width / r_uiscale->value;
		d3d_2DMode.Height = (float) viddef.Height / r_uiscale->value;
	}
	else
	{
		d3d_2DMode.Width = viddef.Width;
		d3d_2DMode.Height = viddef.Height;
	}

	R_BeginFrame (0);

	if (scr_draw_loading == 2)
	{
		//  loading plaque over black screen
		int	w, h;

		R_SetPalette (NULL);
		Draw_Fill (0, 0, d3d_2DMode.Width, d3d_2DMode.Height, 0);
		scr_draw_loading = 0;
		Draw_GetPicSize (&w, &h, "loading");
		Draw_Pic ((d3d_2DMode.Width - w) / 2, (d3d_2DMode.Height - h) / 2, "loading");
	}
	else if (cl.cinematictime > 0)
	{
		// if a cinematic is supposed to be running, handle menus
		// and console specially
		if (cls.key_dest == key_menu)
		{
			SCR_DisableCinematicPalette ();
			M_Draw ();
		}
		else if (cls.key_dest == key_console)
		{
			SCR_DisableCinematicPalette ();
			SCR_DrawConsole ();
		}
		else SCR_DrawCinematic ();
	}
	else
	{
		// make sure the game palette is active
		SCR_DisableCinematicPalette ();

		// do 3D refresh drawing, and then update the screen
		SCR_CalcVrect ();

		// clear any dirty part of the background
		SCR_TileClear ();

		V_RenderView (0);

		SCR_DrawStats ();

		if (cl.frame.playerstate.stats[STAT_LAYOUTS] & 1) SCR_DrawLayout ();
		if (cl.frame.playerstate.stats[STAT_LAYOUTS] & 2) CL_DrawInventory ();

		SCR_DrawNet ();
		SCR_CheckDrawCenterString ();

		if (scr_timegraph->value) SCR_DebugGraph (cls.frametime * 300, 0);
		if (scr_debuggraph->value || scr_timegraph->value || scr_netgraph->value) SCR_DrawDebugGraph ();

		SCR_DrawPause ();

		SCR_DrawConsole ();

		M_Draw ();

		SCR_DrawLoading ();
	}

	GLimp_EndFrame ();
}
