
#include "qcommon.h"

#include <windows.h>
#include <map>


// this buffer should never be accessed directly but always via a call to Scratch_Alloc
static char scratchbuf[20 * 1024 * 1024];
int scratch_base = 0;

// redo this as a virtualalloc buffer...
void *Scratch_Alloc (void)
{
	return scratchbuf;
}


void *Scratch_Alloc (int size)
{
	byte *data = (byte *) Scratch_Alloc ();

	data += scratch_base;
	scratch_base += size;

	return data;
}


void Scratch_Free (void)
{
	scratch_base = 0;
}


int Scratch_LowMark (void)
{
	return scratch_base;
}


void Scratch_FreeToLowMark (int mark)
{
	scratch_base = mark;
}


// wrap the heap object so that we can safely construct
QHEAP::QHEAP (void)
{
	this->hHeap = HeapCreate (0, 0, 0);
}

QHEAP::QHEAP (const QHEAP &other)
{
	this->hHeap = other.hHeap;
}

QHEAP &QHEAP::operator= (const QHEAP &other)
{
	this->hHeap = other.hHeap;
	return *this;
}

QHEAP::~QHEAP (void)
{
	// we can't destruct this normally as the destructor can also run during map traversal.
	// that's OK however as everything is going to be released when the process exits anyway.
	// we could probably refcount it sometime, but right now it doesn't seem such a big deal
}

void *QHEAP::Alloc (int size)
{
	if (size < 1) size = 1;

	if (!this->hHeap)
		this->hHeap = HeapCreate (0, 0, 0);

	return HeapAlloc (this->hHeap, 0, size);
}

void QHEAP::Free (void *ptr)
{
	if (!ptr) return;
	if (!this->hHeap) return;

	HeapFree (this->hHeap, 0, ptr);
}

void QHEAP::Destroy (void)
{
	if (!this->hHeap) return;

	HeapDestroy (this->hHeap);
	this->hHeap = NULL;
}


/*
==============================================================================

						ZONE MEMORY ALLOCATION

just cleared malloc with counters now...

writing your own malloc and free around this is cutesy but is it a worthwhile
waste of your time?  HeapAlloc, HeapFree and (especially) HeapDestroy just do
so much better here.

this also resolves a crasher when trasitioning from biggun

==============================================================================
*/

// yayyy!!! STL!!!!
static std::map<int, QHEAP> zonehandles;

#define	Z_MAGIC		0x1d1d1d1d

struct zhead_t
{
	int		magic;
	int		tag;
};

void Z_Free (void *ptr)
{
	if (ptr)
	{
		zhead_t	*z = ((zhead_t *) ptr) - 1;

		if (z->magic != Z_MAGIC)
			Com_Error (ERR_FATAL, "Z_Free: bad magic");

		zonehandles[z->tag].Free (z);
	}
}


void *Z_TagMalloc (int size, int tag)
{
	size = size + sizeof (zhead_t);

	zhead_t	*z = (zhead_t *) zonehandles[tag].Alloc (size);

	memset (z, 0, size);

	z->magic = Z_MAGIC;
	z->tag = tag;

	return (void *) (z + 1);
}


void Z_FreeTags (int tag)
{
	zonehandles[tag].Destroy ();
}


void *Z_Malloc (int size)
{
	return Z_TagMalloc (size, 0);
}




//===============================================================================

static int hunkcount;
static byte	*membase;
static int hunkmaxsize;
static int cursize;


void *Hunk_Begin (int maxsize)
{
	// reserve a huge chunk of memory, but don't commit any yet
	cursize = 0;
	hunkmaxsize = maxsize;

	membase = (byte *) VirtualAlloc (NULL, maxsize, MEM_RESERVE, PAGE_NOACCESS);

	if (!membase) Sys_Error ("VirtualAlloc reserve failed");

	return (void *) membase;
}


void *Hunk_Alloc (int size)
{
	// round to cacheline
	size = (size + 31) & ~31;

	if ((cursize + size) > hunkmaxsize)
		Sys_Error ("Hunk_Alloc overflow");

	// commit pages as needed
	void *buf = VirtualAlloc (membase, cursize + size, MEM_COMMIT, PAGE_READWRITE);

	if (!buf)
	{
		FormatMessage (
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			GetLastError (),
			MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &buf,
			0,
			NULL
		);

		Sys_Error ("VirtualAlloc commit failed.\n%s", buf);
	}

	cursize += size;

	return (void *) (membase + cursize - size);
}


int Hunk_End (void)
{
	hunkcount++;
	return cursize;
}


void Hunk_Free (void *base)
{
	if (base)
		VirtualFree (base, 0, MEM_RELEASE);

	hunkcount--;
}


